/* 
 *  DGAVCPulldown Copyright (C) 2009-2010, Donald A. Graft, All Rights Reserved
 *
 *  This file is part of DGAVCPulldown, a 3:2 pulldown utility for AVC video.
 *	
 *  DGAVCPulldown is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DGAVCPulldown is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <windows.h>
#include "resource.h"
#include <stdio.h> 
#include <io.h>
#include <fcntl.h>
#include <commctrl.h>
#include <math.h>
#include <sys\stat.h>
#include "global.h"

static DWORD threadId, HRDthreadId;
static HANDLE threadHandle;
static unsigned long pOldEditProc;

static void HandleDrop(HDROP drop)
{
    int count;
    char buf[MAX_PATH];

    count = DragQueryFile(drop, 0xFFFFFFFF, buf, MAX_PATH);

    if (count > 1)
		MessageBox(hWnd, "Multiple files dropped - only one file will be processed", "Warning", MB_OK | MB_ICONWARNING);

    if (DragQueryFile(drop, 0, buf, MAX_PATH) == 0)
	{
		MessageBox(hWnd, "Unable to get dropped file information", "Error", MB_OK | MB_ICONSTOP);
		return;
    }

    SetWindowText(GetDlgItem(hWnd, IDC_INPUT), buf);
	strcpy_s(input_file, sizeof(input_file), buf);
}

static LRESULT CALLBACK NewEditSubWindowProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    switch (msg)
    {
	case WM_DROPFILES:
	    HandleDrop((HDROP)wp);
	    return TRUE;

	default:
	    return CallWindowProc((WNDPROC)pOldEditProc, hwnd, msg, wp, lp);
    }
}

static BOOL CALLBACK DlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	HICON hicon;

	switch(Message)
	{
		case WM_INITDIALOG:
			hWnd = hwnd;
			hicon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ICON)); 
			SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_SMALL, (LPARAM) hicon); 
			SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_BIG, (LPARAM) hicon); 
			pOldEditProc = SetWindowLong(GetDlgItem(hwnd, IDC_INPUT), GWL_WNDPROC, (LONG)NewEditSubWindowProc);
			EnableWindow(GetDlgItem(hWnd, IDC_ABORT), FALSE);
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_ANALYZE:
					if (threadId && WaitForSingleObject(threadHandle, 0) != WAIT_OBJECT_0)
						break;
					GetDlgItemText(hwnd, IDC_INPUT, input_file, 2048);
					StopFlag = 0;
					threadHandle = CreateThread(NULL, 1024000, process, (LPVOID) 0, 0, &threadId);
					break;

				case IDC_ABORT:
					StopFlag = 1;
					break;

				case IDC_BROWSE:
					{
					OPENFILENAME ofn;
					input_file[0] = 0;
					ofn.lStructSize = sizeof(OPENFILENAME);
					ofn.hwndOwner = hwnd;
					ofn.hInstance = NULL;
					ofn.lpTemplateName = NULL;
					ofn.lpstrFilter = "All Files\0*.*\0\0";
					ofn.lpstrCustomFilter = NULL;
					ofn.nMaxCustFilter = 0;
					ofn.nFilterIndex = 1;
					ofn.lpstrFile = input_file;
					ofn.nMaxFile = 1024;
					ofn.lpstrFileTitle = NULL;
					ofn.nMaxFileTitle = 0;
					ofn.lpstrInitialDir = NULL;
					ofn.lpstrTitle = "Choose Input File";
					ofn.Flags = OFN_CREATEPROMPT;
					ofn.nFileOffset = 0;
					ofn.nFileExtension = 0;
					ofn.lpstrDefExt = NULL;
					ofn.lCustData = 0;
					ofn.lpfnHook = NULL;
					GetOpenFileName(&ofn);
					if (input_file[0] != 0)
					{
						SetWindowText(GetDlgItem(hwnd, IDC_INPUT), input_file);
					}
					break;
					}
			}
			break;
		case WM_CLOSE:
			SetWindowLong(GetDlgItem(hwnd, IDC_INPUT), GWL_WNDPROC, (LONG)pOldEditProc);
			EndDialog(hwnd, 0);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	INITCOMMONCONTROLSEX InitCtrlEx;
	InitCtrlEx.dwSize = sizeof(INITCOMMONCONTROLSEX);
	InitCtrlEx.dwICC  = ICC_PROGRESS_CLASS;
	InitCommonControlsEx(&InitCtrlEx);
	return DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, DlgProc);
}

static char printString[1024];
static int enable_popups = 1;
#define NUM_SUPPRESS_STRINGS 256
static unsigned long suppress_list[NUM_SUPPRESS_STRINGS];
static int suppress_list_index;

void init_message_handling(void)
{
	int i;

	for (i = 0; i < NUM_SUPPRESS_STRINGS; i++)
		suppress_list[i] = 0;
	suppress_list_index = 0;
	// This error always comes after a leading B frame error, so suppress it from the start.
	suppress_list[suppress_list_index++] = 2730968692;
	suppress_list[suppress_list_index++] = 344009147;
}

void disable_all_popups(void)
{
	enable_popups = 0;
}

unsigned long hash(char *p)
{
    unsigned long hash = 5381;
    int c;
	unsigned char *str = (unsigned char*) p;

    while (c = *str++)
        hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

    return hash;
}

int dprintf(const char* fmt, ...)
{
	va_list argp;
	unsigned long hash_value;
	int i, reply;

	if (!enable_popups)
		return 0;
	va_start(argp, fmt);
    vsprintf(printString, fmt, argp);
    va_end(argp);
	hash_value = hash(printString);
	for (i = 0; i < suppress_list_index; i++)
		if (suppress_list[i] == hash_value)
			break;
	if (i >= suppress_list_index)
	{
		strcat(printString, "\nHit Yes to continue notifying this specific error type; hit No to disable it."
			"\nHit Cancel to suppress all further errors.");
		reply = MessageBox(hWnd, printString, "Warning", MB_YESNOCANCEL | MB_ICONWARNING);
		if (reply == IDNO)
		{
			if (suppress_list_index < NUM_SUPPRESS_STRINGS - 1)
				suppress_list[suppress_list_index++] = hash_value;
		}
		else if (reply == IDCANCEL)
		{
			enable_popups = 0;
		}
		return strlen(printString);
	}
	else
		return 0;
}
