#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "global.h"

static unsigned char Buffer[BUFFER_SIZE];
static unsigned char *Rdptr;
static size_t Read;

void InitializeBuffer(void)
{
	Read = fread(Buffer, 1, BUFFER_SIZE, fin);
	Rdptr = Buffer;
	IsFirstByteStreamNALU = 1;
}

static unsigned int GetByte(unsigned char *v)
{
	unsigned int t;

	t = (unsigned int) (Rdptr - Buffer);
	if (t < Read)
	{
		*v = *Rdptr++;
		return 0;
	}
	if (t >= BUFFER_SIZE)
	{
		Read = fread(Buffer, 1, BUFFER_SIZE, fin);
		if (Read == 0)
			return -1;
		Rdptr = Buffer;
		*v = *Rdptr++;
		return 0;
	}
	return -1;
}

int GetAnnexbNALU (NALU_t *nalu, __int64 *start_code_pos)
{
  int pos = 0, rewind;
  int TrailingZero8Bits = 0;
  static int HadEOF;
  static unsigned char nalu_buf[BUFFER_SIZE];

  if (IsFirstByteStreamNALU)
  {
    // For the first NALU, skip any crap that may be at the start of the file,
    // and locate the first start code.
    while(1)
    {
		if (pos >= BUFFER_SIZE - 1)
		{
			dprintf("GetAnnexbNALU can't read start code");
			return -1;
		}
        if (GetByte(&nalu_buf[pos++]) == -1)
			return 0;
        if (pos >= 3 && nalu_buf[pos-3] == 0 && nalu_buf[pos-2] == 0 && nalu_buf[pos-1] == 1)
		{
            break;
		}
		HadEOF = 0;
    }
    pos = 0;
  }
  else if (HadEOF)
	  return EOS;

  nalu->startcodeprefix_len = 3;
  nalu_buf[0] = 0;
  nalu_buf[1] = 0;
  nalu_buf[2] = 1;
  pos = 3;
  IsFirstByteStreamNALU=0;

  while (1)
  {
    if (GetByte(&nalu_buf[pos]) == -1)
    {
      nalu->len = pos;
      nalu->buf = (byte *) &nalu_buf[0];
      nalu->forbidden_bit = (nalu->buf[nalu->startcodeprefix_len]>>7) & 1;
      nalu->nal_reference_idc = (nalu->buf[nalu->startcodeprefix_len]>>5) & 3;
      nalu->nal_unit_type = (nalu->buf[nalu->startcodeprefix_len]) & 0x1f;
	  HadEOF = 1;
      return 1;
    }
	pos++;
    if (nalu_buf[pos-3] == 0 && nalu_buf[pos-2] == 0 && nalu_buf[pos-1] == 1)
		break;
  }
  if (nalu_buf[pos-4] == 0)
  {
	  rewind = 4;
      while(nalu_buf[pos-5-TrailingZero8Bits] == 0)
        TrailingZero8Bits++;
  }
  else
	  rewind = 3;

  nalu->len = pos - rewind - TrailingZero8Bits;
  nalu->buf = (byte *) &nalu_buf[0];
  nalu->forbidden_bit = (nalu->buf[nalu->startcodeprefix_len]>>7) & 1;
  nalu->nal_reference_idc = (nalu->buf[nalu->startcodeprefix_len]>>5) & 3;
  nalu->nal_unit_type = (nalu->buf[nalu->startcodeprefix_len]) & 0x1f;

  return (1);
}

int RBSPtoSODB(byte *streamBuffer, int last_byte_pos)
{
  int ctr_bit, bitoffset;

  bitoffset = 0;
  //find trailing 1
  ctr_bit = (streamBuffer[last_byte_pos-1] & (0x01<<bitoffset));   // set up control bit

  while (ctr_bit==0)
  {                 // find trailing 1 bit
    bitoffset++;
    if(bitoffset == 8)
    {
      if(last_byte_pos == 0)
        dprintf(" Panic: All zero data sequence in RBSP");
      assert(last_byte_pos != 0);
      last_byte_pos -= 1;
      bitoffset = 0;
    }
    ctr_bit= streamBuffer[last_byte_pos-1] & (0x01<<(bitoffset));
  }
  return(last_byte_pos);
}

int EBSPtoRBSP(byte *streamBuffer, int end_bytepos, int begin_bytepos)
{
  int i, j, count;
  count = 0;

  if(end_bytepos < begin_bytepos)
    return end_bytepos;

  j = begin_bytepos;

  for(i = begin_bytepos; i < end_bytepos; i++)
  { //starting from begin_bytepos to avoid header information
    if(count == ZEROBYTES_SHORTSTARTCODE && streamBuffer[i] == 0x03)
    {
	  if (i == end_bytepos - 1)
	  {
	    // Special case where the end of the buffer is 00 00 03.
        // Just discard it.
		return j;
	  }
      i++;
      count = 0;
    }
    streamBuffer[j] = streamBuffer[i];
    if(streamBuffer[i] == 0x00)
      count++;
    else
      count = 0;
    j++;
  }

  return j;
}

static unsigned char new_nalu[BUFFER_SIZE];

int RBSPtoEBSP(byte *streamBuffer, int end_bytepos, int begin_bytepos)
{
	int i, j, added = 0;

	j = 3;
	if (end_bytepos > 3)
	{
		new_nalu[0] = streamBuffer[0];
		new_nalu[1] = streamBuffer[1];
		new_nalu[2] = streamBuffer[2];
		for (i = 3; i < end_bytepos;)
		{
			if (streamBuffer[i-2] == 0 && streamBuffer[i-1] == 0 &&
				(streamBuffer[i] == 0 || streamBuffer[i] == 1 ||
				 streamBuffer[i] == 2 || streamBuffer[i] == 3))
			{
				new_nalu[j++] = 3;
				new_nalu[j++] = streamBuffer[i++];
				streamBuffer[i-2] = 0x03;
				added++;
			}
			else
				new_nalu[j++] = streamBuffer[i++];
		}
		if (streamBuffer[i-1] == 0)
		{
			new_nalu[j] = 0x03;
			added++;
		}
	}
	else
	{
		memcpy(new_nalu, streamBuffer, end_bytepos);
		if (new_nalu[end_bytepos-1] == 0)
		{
			new_nalu[j] = 0x03;
			added++;
		}
	}
	memcpy(streamBuffer, new_nalu, end_bytepos + added);
	return end_bytepos + added;
}
