#include <windows.h>
#include "stdio.h"
#include "global.h"
#include "parset.h"
#include "sei.h"
#include "vlc.h"
#include "assert.h"

void decode_poc(struct img_par *img);

static unsigned char buffer[MAX_CODED_FRAME_SIZE];
static unsigned char sc[4] = {0, 0, 0, 1};

int parse_nalu(NALU_t *lnalu, __int64 ftell_position, int record)
{
    int ret = PICTURE_NOT_YET_FOUND;
    Slice *currSlice = img->currentSlice;
    Bitstream *currStream;
	int is_pic_timing_sei;

	lnalu->len -= lnalu->startcodeprefix_len;
	lnalu->buf += lnalu->startcodeprefix_len;
    lnalu->len = EBSPtoRBSP(lnalu->buf, lnalu->len, 1);

    if (lnalu->forbidden_bit)
    {
		// Skip this bad NALU.
		return ret;
    }

	is_pic_timing_sei = 0;
	switch (lnalu->nal_unit_type)
    {
        case NALU_TYPE_SLICE:
        case NALU_TYPE_IDR:
			if (!HadSPS || !HadPPS)
				break;
            img->idr_flag = (lnalu->nal_unit_type == NALU_TYPE_IDR);
            img->nal_reference_idc = lnalu->nal_reference_idc;
            currSlice->dp_mode = PAR_DP_1;
            currSlice->ei_flag = 0;
            currStream = currSlice->partArr[0].bitstream;
            currStream->ei_flag = 0;
            currStream->frame_bitoffset = currStream->read_len = 0;
			currStream->streamBuffer = &lnalu->buf[1];
			currStream->code_len = currStream->bitstream_length = RBSPtoSODB(currStream->streamBuffer, lnalu->len-1);

            // Some syntax of the Slice Header depends on the parameter set, which depends on
            // the parameter set ID of the Slice header.  Hence, read the pic_parameter_set_id
            // of the slice header first, then setup the active parameter sets, and then read
            // the rest of the slice header
            FirstPartOfSliceHeader();
            UseParameterSet (currSlice->pic_parameter_set_id);
            RestOfSliceHeader ();

            hadpic = 0;
            if ((img->frame_num != prev_frame_num) ||
                (currSlice->pic_parameter_set_id != prev_pic_parameter_set_id) ||
                (img->field_pic_flag != prev_field_pic_flag) ||
                ((prev_nal_ref_idc != img->nal_reference_idc) &&
                 ((prev_nal_ref_idc == 0) || (img->nal_reference_idc == 0)))
                )
            {
                hadpic = 1;
            }
            if(img->field_pic_flag && prev_field_pic_flag)
            {
              if (prev_bottom_field_flag != img->bottom_field_flag)
                  hadpic = 1;
            }
            if (prev_idr_flag != img->idr_flag)
                hadpic = 1;
            if (active_sps->pic_order_cnt_type == 0)
            {
              if (prev_pic_order_cnt_lsb != img->pic_order_cnt_lsb)
                  hadpic = 1;
              if (prev_delta_pic_order_cnt_bottom != img->delta_pic_order_cnt_bottom)
                  hadpic = 1;
            }
            if (active_sps->pic_order_cnt_type == 1)
            {
              if (prev_delta_pic_order_cnt0 != img->delta_pic_order_cnt[0])
                  hadpic = 1;
              if (prev_delta_pic_order_cnt1 != img->delta_pic_order_cnt[1])
                  hadpic = 1;
            }
			if (prev_idr_flag == 1 && img->idr_flag == 1 && (img->idr_pic_id != prev_idr_pic_id))
				hadpic = 1;

            prev_frame_num = img->frame_num;
            prev_field_pic_flag = img->field_pic_flag;
            prev_bottom_field_flag = img->bottom_field_flag;
            prev_pic_parameter_set_id = currSlice->pic_parameter_set_id;
            prev_nal_ref_idc = img->nal_reference_idc;
            prev_idr_flag = img->idr_flag;
            prev_idr_pic_id = img->idr_pic_id;
            prev_delta_pic_order_cnt0 = img->delta_pic_order_cnt[0];
            prev_delta_pic_order_cnt1 = img->delta_pic_order_cnt[1];
            prev_pic_order_cnt_lsb = img->pic_order_cnt_lsb;
            prev_delta_pic_order_cnt_bottom = img->delta_pic_order_cnt_bottom;

			if (hadpic && (!img->field_pic_flag || (img->field_pic_flag && !img->bottom_field_flag)))
            {
				int poc;

				decode_poc(img);
				poc = img->framepoc / 2;
				pocs[pocs_ndx++] = poc;
				ret = PICTURE_FOUND;
				break;
            }
			break;
	  case NALU_TYPE_DPA:
        dprintf ("data partitioning not implemented\n");
        break;
      case NALU_TYPE_DPB:
        dprintf ("found data partition B without matching DP A, discarding\n");
        break;
      case NALU_TYPE_DPC:
        dprintf ("found data partition C without matching DP A, discarding\n");
        break;
      case NALU_TYPE_SEI:
		is_pic_timing_sei = InterpretSEIMessage(lnalu->buf,lnalu->len, img, record);
        break;
      case NALU_TYPE_PPS:
	    HadPPS = 1;
        ProcessPPS(lnalu);
		break;
      case NALU_TYPE_SPS:
		HadSPS = 1;
        ProcessSPS(lnalu);
        break;
      case NALU_TYPE_AUD:
        break;
      case NALU_TYPE_EOSEQ:
        break;
      case NALU_TYPE_EOSTREAM:
        ret = EOS;
		break;
      case NALU_TYPE_FILL:
        break;
      default:
        dprintf ("Found NALU type %d, len %d undefined, ignore NALU, moving on\n", lnalu->nal_unit_type, lnalu->len);
		break;
    }
	if (operation == OP_PROCESS)
	{
		if (had_pic_sei && is_pic_timing_sei && new_sei[0])
		{
			// We edited the existing SEI.
			NALU_t mnalu;

			mnalu.buf = new_sei;
			mnalu.len = new_sei[2] + 4;
			mnalu.len = RBSPtoEBSP(mnalu.buf, mnalu.len, 1);
			fwrite(sc, 1, 4, fout);
			fwrite(mnalu.buf, 1, mnalu.len, fout);
		}
		else if (!had_pic_sei && ret == PICTURE_FOUND)
		{
			unsigned char sei[9];

			if (had_hrd)
			{
				MessageBox(hWnd, "This combination not yet supported -- contact neuron2!", "Error", MB_OK | MB_ICONSTOP);
				ExitThread(0);
			}

			// Have to generate a pic timing SEI.
			sei[0] = 0;
			sei[1] = 0;
			sei[2] = 0;
			sei[3] = 1;
			sei[4] = 6;
			sei[5] = 1;
			sei[6] = 1;
			switch (rpts[sei_32_ndx++])
			{
			case 5:
				sei[7] = 0x51;
				break;
			case 4:
				sei[7] = 0x42;
				break;
			case 6:
				sei[7] = 0x61;
				break;
			case 3:
				sei[7] = 0x32;
				break;
			}
			sei[8] = 0x80;
			fwrite(sei, 1, 9, fout);
			lnalu->len = RBSPtoEBSP(lnalu->buf, lnalu->len, 1);
			fwrite(sc, 1, 4, fout);
			fwrite(lnalu->buf, 1, lnalu->len, fout);
		}
		else
		{
			lnalu->len = RBSPtoEBSP(lnalu->buf, lnalu->len, 1);
			fwrite(sc, 1, 4, fout);
			fwrite(lnalu->buf, 1, lnalu->len, fout);
		}
	}
	return ret;
}

int read_new_picture()
{
    __int64 ftell_position;
	int ret;

	while (1)
    {
		ret = GetAnnexbNALU (nalu, &ftell_position);
		if (ret < 0)
			dprintf ("Error while getting the Annex B NALU, exit\n");
		if (ret == EOS)
			break;
		bytes_delivered += nalu->len;
		if (nalu->len == 0)
			continue;
		if ((ret = parse_nalu(nalu, ftell_position, 0)) != PICTURE_NOT_YET_FOUND)
			break;
	}
	return ret;
}

void no_mem_exit(char *where)
{
   dprintf("Could not allocate memory: %s", where);
}

void init_indexing(void)
{
	nalu = &nalu_data;
	img = &img_parameters;
	img->currentSlice = &slice;
	img->currentSlice->partArr = &partition;
	img->currentSlice->partArr->bitstream = &stream;
	img->PrevPicOrderCntMsb = 0;
	img->PrevPicOrderCntLsb = 0;

	active_sps = NULL;
	IsFirstByteStreamNALU = 1;
	HadSPS = HadPPS = 0;
	sei_32[0] = 5;
	sei_32[1] = 4;
	sei_32[2] = 6;
	sei_32[3] = 3;
	sei_32_ndx = 0;

	prev_frame_num = 0xffffffff;
	prev_field_pic_flag = 0xffffffff;
	prev_bottom_field_flag = 0xffffffff;
	prev_pic_parameter_set_id = -1;
	prev_nal_ref_idc = -1;
	prev_idr_flag = -1;
	prev_idr_pic_id = -1;
	prev_delta_pic_order_cnt0 = -1;
	prev_delta_pic_order_cnt1 = -1;
	prev_pic_order_cnt_lsb = 0xffffffff;
	prev_delta_pic_order_cnt_bottom = -1;
}

/*!
 ************************************************************************
 * \brief
 *    read the first part of the header (only the pic_parameter_set_id)
 * \return
 *    Length of the first part of the slice header (in bits)
 ************************************************************************
 */
int FirstPartOfSliceHeader()
{
  Slice *currSlice = img->currentSlice;
  int dP_nr = 0;
  DataPartition *partition = &(currSlice->partArr[dP_nr]);
  Bitstream *currStream = partition->bitstream;
  int tmp;

  // Get first_mb_in_slice
  currSlice->start_mb_nr = ue_v ("SH: first_mb_in_slice", currStream);

  tmp = ue_v ("SH: slice_type", currStream);

  img->type_raw = (SliceType) tmp;
  if (tmp>4) tmp -=5;
  img->type = currSlice->picture_type = (SliceType) tmp;

  currSlice->pic_parameter_set_id = ue_v ("SH: pic_parameter_set_id", currStream);

  return 0;
}

/*!
 ************************************************************************
 * \brief
 *    read the scond part of the header (without the pic_parameter_set_id
 * \return
 *    Length of the second part of the Slice header in bits
 ************************************************************************
 */
int RestOfSliceHeader()
{
  Slice *currSlice = img->currentSlice;
  int dP_nr = 0;
  DataPartition *partition = &(currSlice->partArr[dP_nr]);
  Bitstream *currStream = partition->bitstream;

  img->frame_num = u_v (active_sps->log2_max_frame_num_minus4 + 4, "SH: frame_num", currStream);

  /* Tian Dong: frame_num gap processing, if found */
  if (img->idr_flag)
  {
    img->pre_frame_num = img->frame_num;
    // picture error concealment
    img->last_ref_pic_poc = 0;
    assert(img->frame_num == 0);
  }

  if (active_sps->frame_mbs_only_flag)
  {
    img->structure = FRAME;
    img->field_pic_flag=0;
  }
  else
  {
    // field_pic_flag   u(1)
    img->field_pic_flag = u_1("SH: field_pic_flag", currStream);
    if (img->field_pic_flag)
    {
      // bottom_field_flag  u(1)
      img->bottom_field_flag = u_1("SH: bottom_field_flag", currStream);

      img->structure = img->bottom_field_flag ? BOTTOM_FIELD : TOP_FIELD;
    }
    else
    {
      img->structure = FRAME;
      img->bottom_field_flag=0;
    }
  }

  currSlice->structure = (PictureStructure) img->structure;

  img->MbaffFrameFlag=(active_sps->mb_adaptive_frame_field_flag && (img->field_pic_flag==0));

  if (img->structure == FRAME       ) assert (img->field_pic_flag == 0);
  if (img->structure == TOP_FIELD   ) assert (img->field_pic_flag == 1 && img->bottom_field_flag == 0);
  if (img->structure == BOTTOM_FIELD) assert (img->field_pic_flag == 1 && img->bottom_field_flag == 1);

  if (img->idr_flag)
  {
    img->idr_pic_id = ue_v("SH: idr_pic_id", currStream);
  }

  if (active_sps->pic_order_cnt_type == 0)
  {
    img->pic_order_cnt_lsb = u_v(active_sps->log2_max_pic_order_cnt_lsb_minus4 + 4, "SH: pic_order_cnt_lsb", currStream);
    if( active_pps->pic_order_present_flag  ==  1 &&  !img->field_pic_flag )
      img->delta_pic_order_cnt_bottom = se_v("SH: delta_pic_order_cnt_bottom", currStream);
    else
      img->delta_pic_order_cnt_bottom = 0;
  }
  if( active_sps->pic_order_cnt_type == 1 && !active_sps->delta_pic_order_always_zero_flag )
  {
    img->delta_pic_order_cnt[ 0 ] = se_v("SH: delta_pic_order_cnt[0]", currStream);
    if( active_pps->pic_order_present_flag  ==  1  &&  !img->field_pic_flag )
      img->delta_pic_order_cnt[ 1 ] = se_v("SH: delta_pic_order_cnt[1]", currStream);
  }else
  {
    if (active_sps->pic_order_cnt_type == 1)
    {
      img->delta_pic_order_cnt[ 0 ] = 0;
      img->delta_pic_order_cnt[ 1 ] = 0;
    }
  }
  return 0;
}

void decode_poc(struct img_par *img)
{
  int i;

  // for POC mode 0:
  unsigned int MaxPicOrderCntLsb = (1<<(active_sps->log2_max_pic_order_cnt_lsb_minus4+4));

  switch ( active_sps->pic_order_cnt_type )
  {
  case 0: // POC MODE 0
    // 1st
    if(img->idr_flag)
    {
      img->PrevPicOrderCntMsb = 0;
      img->PrevPicOrderCntLsb = 0;
    }
    else
    {
      if (img->last_has_mmco_5)
      {
        if (img->last_pic_bottom_field)
        {
          img->PrevPicOrderCntMsb = 0;
          img->PrevPicOrderCntLsb = 0;
        }
        else
        {
          img->PrevPicOrderCntMsb = 0;
          img->PrevPicOrderCntLsb = img->toppoc;
        }
      }
    }
    // Calculate the MSBs of current picture
    if( img->pic_order_cnt_lsb  <  img->PrevPicOrderCntLsb  &&
      ( img->PrevPicOrderCntLsb - img->pic_order_cnt_lsb )  >=  ( MaxPicOrderCntLsb / 2 ) )
      img->PicOrderCntMsb = img->PrevPicOrderCntMsb + MaxPicOrderCntLsb;
    else if ( img->pic_order_cnt_lsb  >  img->PrevPicOrderCntLsb  &&
      ( img->pic_order_cnt_lsb - img->PrevPicOrderCntLsb )  >  ( MaxPicOrderCntLsb / 2 ) )
      img->PicOrderCntMsb = img->PrevPicOrderCntMsb - MaxPicOrderCntLsb;
    else
      img->PicOrderCntMsb = img->PrevPicOrderCntMsb;

    // 2nd

    if(img->field_pic_flag==0)
    {           //frame pix
      img->toppoc = img->PicOrderCntMsb + img->pic_order_cnt_lsb;
      img->bottompoc = img->toppoc + img->delta_pic_order_cnt_bottom;
      img->ThisPOC = (img->toppoc < img->bottompoc)? img->toppoc : img->bottompoc; // POC200301
    }
    else if (img->bottom_field_flag==0)
    {  //top field
      img->ThisPOC= img->toppoc = img->PicOrderCntMsb + img->pic_order_cnt_lsb;
      }
      else
    {  //bottom field
      img->ThisPOC= img->bottompoc = img->PicOrderCntMsb + img->pic_order_cnt_lsb;
    }
    img->framepoc = img->ThisPOC;

    if ( img->frame_num!=img->PreviousFrameNum)
      img->PreviousFrameNum=img->frame_num;

    if(img->nal_reference_idc)
    {
      img->PrevPicOrderCntLsb = img->pic_order_cnt_lsb;
      img->PrevPicOrderCntMsb = img->PicOrderCntMsb;
    }

    break;

  case 1: // POC MODE 1
    // 1st
    if(img->idr_flag)
    {
      img->FrameNumOffset=0;     //  first pix of IDRGOP,
      img->delta_pic_order_cnt[0]=0;                        //ignore first delta
      if(img->frame_num)
        dprintf("frame_num not equal to zero in IDR picture");
    }
    else
    {
      if (img->last_has_mmco_5)
      {
        img->PreviousFrameNumOffset = 0;
        img->PreviousFrameNum = 0;
      }
      if (img->frame_num<img->PreviousFrameNum)
      {             //not first pix of IDRGOP
        img->FrameNumOffset = img->PreviousFrameNumOffset + img->MaxFrameNum;
      }
      else
      {
        img->FrameNumOffset = img->PreviousFrameNumOffset;
      }
    }

    // 2nd
    if(active_sps->num_ref_frames_in_pic_order_cnt_cycle)
      img->AbsFrameNum = img->FrameNumOffset+img->frame_num;
    else
      img->AbsFrameNum=0;
    if( (!img->nal_reference_idc) && img->AbsFrameNum>0)
      img->AbsFrameNum--;

    // 3rd
    img->ExpectedDeltaPerPicOrderCntCycle=0;

    if(active_sps->num_ref_frames_in_pic_order_cnt_cycle)
    for(i=0;i<(int) active_sps->num_ref_frames_in_pic_order_cnt_cycle;i++)
      img->ExpectedDeltaPerPicOrderCntCycle += active_sps->offset_for_ref_frame[i];

    if(img->AbsFrameNum)
    {
      img->PicOrderCntCycleCnt = (img->AbsFrameNum-1)/active_sps->num_ref_frames_in_pic_order_cnt_cycle;
      img->FrameNumInPicOrderCntCycle = (img->AbsFrameNum-1)%active_sps->num_ref_frames_in_pic_order_cnt_cycle;
      img->ExpectedPicOrderCnt = img->PicOrderCntCycleCnt*img->ExpectedDeltaPerPicOrderCntCycle;
      for(i=0;i<=(int)img->FrameNumInPicOrderCntCycle;i++)
        img->ExpectedPicOrderCnt += active_sps->offset_for_ref_frame[i];
    }
    else
      img->ExpectedPicOrderCnt=0;

    if(!img->nal_reference_idc)
      img->ExpectedPicOrderCnt += active_sps->offset_for_non_ref_pic;

    if(img->field_pic_flag==0)
    {           //frame pix
      img->toppoc = img->ExpectedPicOrderCnt + img->delta_pic_order_cnt[0];
      img->bottompoc = img->toppoc + active_sps->offset_for_top_to_bottom_field + img->delta_pic_order_cnt[1];
      img->ThisPOC = img->framepoc = (img->toppoc < img->bottompoc)? img->toppoc : img->bottompoc; // POC200301
    }
    else if (img->bottom_field_flag==0)
    {  //top field
      img->ThisPOC = img->toppoc = img->ExpectedPicOrderCnt + img->delta_pic_order_cnt[0];
    }
    else
    {  //bottom field
      img->ThisPOC = img->bottompoc = img->ExpectedPicOrderCnt + active_sps->offset_for_top_to_bottom_field + img->delta_pic_order_cnt[0];
    }
    img->framepoc=img->ThisPOC;

    img->PreviousFrameNum=img->frame_num;
    img->PreviousFrameNumOffset=img->FrameNumOffset;

    break;

  case 2: // POC MODE 2
    if(img->idr_flag) // IDR picture
    {
      img->FrameNumOffset=0;     //  first pix of IDRGOP,
      img->ThisPOC = img->framepoc = img->toppoc = img->bottompoc = 0;
      if(img->frame_num)
        dprintf("frame_num not equal to zero in IDR picture");
    }
    else
    {
      if (img->last_has_mmco_5)
      {
        img->PreviousFrameNum = 0;
        img->PreviousFrameNumOffset = 0;
      }
      if (img->frame_num<img->PreviousFrameNum)
        img->FrameNumOffset = img->PreviousFrameNumOffset + img->MaxFrameNum;
      else
        img->FrameNumOffset = img->PreviousFrameNumOffset;


      img->AbsFrameNum = img->FrameNumOffset+img->frame_num;
      if(!img->nal_reference_idc)
        img->ThisPOC = (2*img->AbsFrameNum - 1);
      else
        img->ThisPOC = (2*img->AbsFrameNum);

      if (img->field_pic_flag==0)
        img->toppoc = img->bottompoc = img->framepoc = img->ThisPOC;
      else if (img->bottom_field_flag==0)
         img->toppoc = img->framepoc = img->ThisPOC;
      else img->bottompoc = img->framepoc = img->ThisPOC;
    }

    img->PreviousFrameNum=img->frame_num;
    img->PreviousFrameNumOffset=img->FrameNumOffset;
    break;

  default:
	dprintf ("Unknown POC order count type!\n");
    break;
  }
}
