/* 
 *  DGAVCPulldown Copyright (C) 2009-2010, Donald A. Graft, All Rights Reserved
 *
 *  This file is part of DGAVCPulldown, a 3:2 pulldown utility for AVC video.
 *	
 *  DGAVCPulldown is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DGAVCPulldown is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <windows.h>
#include "stdlib.h"
#include "stdio.h"
#include <io.h>
#include <fcntl.h>
#include <commctrl.h>
#include <math.h>
#include <sys\stat.h>
#include "string.h"
#include "global.h"
#include "resource.h"

void insertion_sort(int pocs[], int rpts[], int length)
{
  int key, keyr, i, j;

  for(j = 1; j < length; j++)
  {
     key = pocs[j];
     keyr = rpts[j];
     i = j - 1;
     while(pocs[i] > key && i >= 0)
     {
         pocs[i+1] = pocs[i];
         rpts[i+1] = rpts[i];
         i--;
     }
     pocs[i+1] = key;
     rpts[i+1] = keyr;
  }
}

DWORD WINAPI process(LPVOID op)
{
	unsigned __int64 size;
	unsigned __int64 total;
	int fd;
	char *p;
	int i, j, start, end;

	// First analyze the source video.
	operation = OP_ANALYZE;
	// We will set these variables which will guide us in processing.
	had_hrd = 0;
	had_pic_sei = 0;
	had_pic_struct = 0;

	// Open the input file for reading.
	fopen_s(&fin, input_file, "rb");
	if (fin == NULL)
	{
		MessageBox(hWnd, "Cannot open input file", "Error", MB_OK | MB_ICONSTOP);
		ExitThread(0);
	}

	// Get the file size.
	fd = _open(input_file, _O_RDONLY | _O_BINARY | _O_SEQUENTIAL);
	size = _lseeki64(fd, 0, SEEK_END);
	_close(fd);

	// Analysis pass to get stream info and all the POCs.
	SetWindowText(GetDlgItem(hWnd, IDC_STATUS), "Pass 1: analyzing...");
	EnableWindow(GetDlgItem(hWnd, IDC_ANALYZE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_BROWSE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_ABORT), TRUE);

	init_indexing();
	InitializeBuffer();
	total = 0;
	pocs_ndx = 0;
	while (1)
	{
		if (StopFlag)
		{
			SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);
			break;
		}
		bytes_delivered = 0;
		if (read_new_picture() == EOS)
			break;
		total += bytes_delivered;
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, (UINT) (total * 100.0 / size), 0);
	}

	// Done analyzing the input file. Now
	// create the repeats array that will be used to get the picture structure
	// value for the SEIs. We first create the pattern in display order.
	for (i = 0; i < pocs_ndx; i++)
	{
		rpts2[i] = sei_32[i%4];
	}
	// Now we sort each GOP by POC, moving the repeat value along with the
	// POC. This puts the repeat values in encode order for use during the
	// processing pass below.
	end = -1;
	while (1)
	{
		start = end + 1;
		for (i = start + 1; i < pocs_ndx; i++)
		{
			if (pocs[i] == 0)
				break;
		}
		end = i - 1;
		for (j = 0; j < i; j++)
		{
			rpts[start+j] = rpts2[start+pocs[start+j]];
		}
#if 0
        // Sort the GOP.
		if (end > start)
			insertion_sort(&pocs[start], &rpts[start], end - start + 1);
#endif
		if (end >= pocs_ndx - 1)
			break;
	}

	// Now process the input file.
	operation = OP_PROCESS;

	// Open the input file for reading.
	fopen_s(&fin, input_file, "rb");
	if (fin == NULL)
	{
		MessageBox(hWnd, "Cannot open input file", "Error", MB_OK | MB_ICONSTOP);
		ExitThread(0);
	}
	strcpy(output_file, input_file);
	p = output_file + strlen(output_file);
	while (*p != '.') p--;
	*p = 0;
	strcat(p, ".pulldown.264");
	fopen_s(&fout, output_file, "wb");
	if (fout == NULL)
	{
		MessageBox(hWnd, "Cannot open output file", "Error", MB_OK | MB_ICONSTOP);
		ExitThread(0);
	}

	SetWindowText(GetDlgItem(hWnd, IDC_STATUS), "Pass 2: working...");
	EnableWindow(GetDlgItem(hWnd, IDC_ANALYZE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_BROWSE), FALSE);
	EnableWindow(GetDlgItem(hWnd, IDC_ABORT), TRUE);
	
	init_indexing();
	InitializeBuffer();
	total = 0;
	while (1)
	{
		if (StopFlag)
		{
			SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);
			break;
		}
		bytes_delivered = 0;
		if (read_new_picture() == EOS)
			break;
		total += bytes_delivered;
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, (UINT) (total * 100.0 / size), 0);
	}
	fclose(fin);
	fclose(fout);
	SetWindowText(GetDlgItem(hWnd, IDC_STATUS), "Done!");
	EnableWindow(GetDlgItem(hWnd, IDC_ANALYZE), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_BROWSE), TRUE);
	EnableWindow(GetDlgItem(hWnd, IDC_ABORT), FALSE);
	ExitThread(0);
}
