#include <stdlib.h>
#include <assert.h>
#include <string.h>

#define JMGLOBAL
#include "global.h"
#include "sei.h"
#include "vlc.h"
#include "parset.h"

int InterpretSEIMessage(byte* msg, int size, ImageParameters *img, int record)
{
  int payload_type = 0;
  int payload_size = 0;
  int ndx = 1;
  int is_pic_timing_sei = 0;
  byte tmp_byte;
  do
  {
    // sei_message();
    payload_type = 0;
    if (ndx >= size) return 0;
    tmp_byte = msg[ndx++];
    while (tmp_byte == 0xFF)
    {
      payload_type += 255;
      if (ndx >= size) return 0;
      tmp_byte = msg[ndx++];
    }
    payload_type += tmp_byte;   // this is the last byte

    payload_size = 0;
    if (ndx >= size) return 0;
    tmp_byte = msg[ndx++];
    while (tmp_byte == 0xFF)
    {
      payload_size += 255;
      if (ndx >= size) return 0;
      tmp_byte = msg[ndx++];
    }
    payload_size += tmp_byte;   // this is the last byte

    switch ( payload_type )     // sei_payload( type, size );
    {
    case  SEI_PIC_TIMING:
	  if (ndx + payload_size >= size) return 0;
      interpret_picture_timing_info( msg+ndx, payload_size, img, record);
	  is_pic_timing_sei = 1;
	  break;
    case  SEI_RECOVERY_POINT:
	  if (ndx + payload_size >= size) return 0;
      interpret_recovery_point_info( msg+ndx, payload_size, img );
      break;
    default:
      break;
    }
    ndx += payload_size;
    if (ndx >= size) return 0;
  } while( msg[ndx] != 0x80 );

  if (ndx >= size) return 0;
  return is_pic_timing_sei;
}

void interpret_recovery_point_info( byte* payload, int size, ImageParameters *img )
{
  int recovery_frame_cnt, exact_match_flag, broken_link_flag, changing_slice_group_idc;


  Bitstream bitstream;
  Bitstream* buf = &bitstream;

  buf->bitstream_length = size;
  buf->streamBuffer = payload;
  buf->frame_bitoffset = 0;

  recovery_frame_cnt       = ue_v(    "SEI: recovery_frame_cnt"      , buf);
  exact_match_flag         = u_1 (    "SEI: exact_match_flag"        , buf);
  broken_link_flag         = u_1 (    "SEI: broken_link_flag"        , buf);
  changing_slice_group_idc = u_v ( 2, "SEI: changing_slice_group_idc", buf);

  img->recovery_frame_cnt = recovery_frame_cnt;
}

void interpret_picture_timing_info( byte* payload, int size, ImageParameters *img, int record )
{
  int cpb_removal_delay, dpb_output_delay, picture_structure_present_flag, picture_structure;
  int clock_time_stamp_flag;
  int ct_type, nuit_field_based_flag, counting_type, full_timestamp_flag, discontinuity_flag, cnt_dropped_flag, nframes;
  int seconds_value, minutes_value, hours_value, seconds_flag, minutes_flag, hours_flag, time_offset;
  int NumClockTs = 0;
  int i;

  int cpb_removal_len = 24;
  int dpb_output_len  = 24;

  Boolean CpbDpbDelaysPresentFlag;

  Bitstream bitstream;
  Bitstream* buf = &bitstream;

  if (NULL==active_sps)
  {
	  extern seq_parameter_set_rbsp_t SeqParSet[];
	  if (SeqParSet[0].Valid)
	  {
		UseParameterSet(0);
	  }
	  else
	  {
        fprintf (stderr, "Warning: no active SPS, timing SEI cannot be parsed\n");
        return;
	  }
  }

  if (operation == OP_ANALYZE)
	  had_pic_sei = 1;
  buf->bitstream_length = size;
  buf->streamBuffer = payload;
  buf->frame_bitoffset = 0;

  // CpbDpbDelaysPresentFlag can also be set "by some means not specified in this Recommendation | International Standard"
  CpbDpbDelaysPresentFlag =  (Boolean) (active_sps->vui_parameters_present_flag
                              && (   (active_sps->vui_seq_parameters.nal_hrd_parameters_present_flag != 0)
                                   ||(active_sps->vui_seq_parameters.vcl_hrd_parameters_present_flag != 0)));

  if (CpbDpbDelaysPresentFlag )
  {
    if (active_sps->vui_parameters_present_flag)
    {
      if (active_sps->vui_seq_parameters.nal_hrd_parameters_present_flag)
      {
        cpb_removal_len = active_sps->vui_seq_parameters.nal_hrd_parameters.cpb_removal_delay_length_minus1 + 1;
        dpb_output_len  = active_sps->vui_seq_parameters.nal_hrd_parameters.dpb_output_delay_length_minus1  + 1;
      }
      else if (active_sps->vui_seq_parameters.vcl_hrd_parameters_present_flag)
      {
        cpb_removal_len = active_sps->vui_seq_parameters.vcl_hrd_parameters.cpb_removal_delay_length_minus1 + 1;
        dpb_output_len  = active_sps->vui_seq_parameters.vcl_hrd_parameters.dpb_output_delay_length_minus1  + 1;
      }
    }

    if ((active_sps->vui_seq_parameters.nal_hrd_parameters_present_flag)||
      (active_sps->vui_seq_parameters.vcl_hrd_parameters_present_flag))
    {
      cpb_removal_delay = u_v(cpb_removal_len, "SEI: cpb_removal_delay" , buf);
      dpb_output_delay  = u_v(dpb_output_len,  "SEI: dpb_output_delay"  , buf);
    }
  }

  if (!active_sps->vui_parameters_present_flag)
  {
    picture_structure_present_flag = 0;
  }
  else
  {
    picture_structure_present_flag = active_sps->vui_seq_parameters.pic_struct_present_flag;
  }

  if (picture_structure_present_flag)
  {
	if (operation == OP_PROCESS)
	{
		if (had_pic_sei && !had_pic_struct)
		{
			// Adding pic_struct to existing SEI with existing HRD info.
			int clkts, picstruct;

			picstruct = rpts[sei_32_ndx++];
			if (picstruct == 5 || picstruct == 6)
				clkts = 3;
			else
				clkts = 2;
			memset(new_sei, 0, sizeof(new_sei));
			new_sei[0] = 6;
			new_sei[1] = 1;
			new_sei[2] = buf->bitstream_length;
			memcpy(&new_sei[3], buf->streamBuffer, buf->bitstream_length); 
			if ((buf->frame_bitoffset % 8) == 0)
			{
				// Need to add a byte.
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length + 1, 4);
				if (clkts == 2)
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x2, buf->bitstream_length + 1, 4);
				else
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x1, buf->bitstream_length + 1, 4);
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
			else if ((buf->frame_bitoffset % 8) == 1)
			{
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length, 4);
				if (clkts == 2)
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x1, buf->bitstream_length, 3);
				else
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x0, buf->bitstream_length, 3);
			}
			else if ((buf->frame_bitoffset % 8) == 2)
			{
				if (clkts == 2)
				{
					PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length, 4);
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x0, buf->bitstream_length, 2);
				}
				else
				{
					new_sei[2]++;
					PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length + 1, 4);
					PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x0, buf->bitstream_length + 1, 2);
					new_sei[3 + buf->bitstream_length] = 0x40;
					new_sei[3 + buf->bitstream_length + 1] = 0x80;
				}
			}
			else if ((buf->frame_bitoffset % 8) == 3)
			{
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length + 1, 4);
				PutBits(&new_sei[3], buf->frame_bitoffset + 4, 0x0, buf->bitstream_length + 1, 1);
				if (clkts == 2)
					new_sei[3 + buf->bitstream_length] = 0x40;
				else
					new_sei[3 + buf->bitstream_length] = 0x20;
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
			else if ((buf->frame_bitoffset % 8) == 4)
			{
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct, buf->bitstream_length + 1, 4);
				if (clkts == 2)
					new_sei[3 + buf->bitstream_length] = 0x20;
				else
					new_sei[3 + buf->bitstream_length] = 0x10;
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
			else if ((buf->frame_bitoffset % 8) == 5)
			{
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct >> 1, buf->bitstream_length + 1, 3);
				if (clkts == 2)
					new_sei[3 + buf->bitstream_length] = (picstruct << 7) | 0x10;
				else
					new_sei[3 + buf->bitstream_length] = (picstruct << 7) | 0x08;
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
			else if ((buf->frame_bitoffset % 8) == 6)
			{
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct >> 2, buf->bitstream_length + 1, 2);
				if (clkts == 2)
					new_sei[3 + buf->bitstream_length] = (picstruct << 6) | 0x08;
				else
					new_sei[3 + buf->bitstream_length] = (picstruct << 6) | 0x04;
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
			else if ((buf->frame_bitoffset % 8) == 7)
			{
				new_sei[2]++;
				PutBits(&new_sei[3], buf->frame_bitoffset, picstruct >> 3, buf->bitstream_length + 1, 1);
				if (clkts == 2)
					new_sei[3 + buf->bitstream_length] = (picstruct << 5) | 0x04;
				else
					new_sei[3 + buf->bitstream_length] = (picstruct << 5) | 0x02;
				new_sei[3 + buf->bitstream_length + 1] = 0x80;
			}
		}
		else if (had_pic_sei && had_pic_struct)
		{
			MessageBox(hWnd, "Stream already defines picture structure!", "Error", MB_OK | MB_ICONSTOP);
			ExitThread(0);
		}
	}
    picture_structure = u_v(4, "SEI: pic_struct" , buf);
    switch (picture_structure)
    {
    case 0:
    case 1:
    case 2:
      NumClockTs = 1;
      break;
    case 3:
    case 4:
      NumClockTs = 2;
      break;
    case 7:
      NumClockTs = 2;
      break;
    case 5:
    case 6:
      NumClockTs = 3;
      break;
    case 8:
      NumClockTs = 3;
      break;
    default:
      dprintf("reserved picture_structure used (can't determine NumClockTs)");
    }
    for (i=0; i<NumClockTs; i++)
    {
      clock_time_stamp_flag = u_1("SEI: clock_time_stamp_flag"  , buf);
      if (clock_time_stamp_flag)
      {
        ct_type               = u_v(2, "SEI: ct_type"               , buf);
        nuit_field_based_flag = u_1(   "SEI: nuit_field_based_flag" , buf);
        counting_type         = u_v(5, "SEI: counting_type"         , buf);
        full_timestamp_flag   = u_1(   "SEI: full_timestamp_flag"   , buf);
        discontinuity_flag    = u_1(   "SEI: discontinuity_flag"    , buf);
        cnt_dropped_flag      = u_1(   "SEI: cnt_dropped_flag"      , buf);
        nframes               = u_v(8, "SEI: nframes"               , buf);

        if (full_timestamp_flag)
        {
          seconds_value         = u_v(6, "SEI: seconds_value"   , buf);
          minutes_value         = u_v(6, "SEI: minutes_value"   , buf);
          hours_value           = u_v(5, "SEI: hours_value"     , buf);
        }
        else
        {
          seconds_flag          = u_1(   "SEI: seconds_flag" , buf);
          if (seconds_flag)
          {
            seconds_value         = u_v(6, "SEI: seconds_value"   , buf);
            minutes_flag          = u_1(   "SEI: minutes_flag" , buf);
            if(minutes_flag)
            {
              minutes_value         = u_v(6, "SEI: minutes_value"   , buf);
              hours_flag            = u_1(   "SEI: hours_flag" , buf);
              if(hours_flag)
              {
                hours_value           = u_v(5, "SEI: hours_value"     , buf);
              }
            }
          }
        }
        {
          int time_offset_length;
          if (active_sps->vui_seq_parameters.vcl_hrd_parameters_present_flag)
            time_offset_length = active_sps->vui_seq_parameters.vcl_hrd_parameters.time_offset_length;
          else if (active_sps->vui_seq_parameters.nal_hrd_parameters_present_flag)
            time_offset_length = active_sps->vui_seq_parameters.nal_hrd_parameters.time_offset_length;
          else
            time_offset_length = 24;
          if (time_offset_length)
            time_offset = u_v(time_offset_length, "SEI: time_offset"   , buf); // TODO interpretation is unsigned, need signed interpretation (i_v)
          else
            time_offset = 0;
        }
      }
    }
  }
}
